@@ -15,6 +15,7 @@
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
+#include <my_global.h>
 #include <my_time.h>
 #include <m_string.h>
 #include <m_ctype.h>
@@ -292,7 +293,7 @@ static void get_microseconds(ulong *val, MYSQL_TIME_STATUS *status,
 #define MAX_DATE_PARTS 8
 
 my_bool
-str_to_datetime(const char *str, uint length, MYSQL_TIME *l_time,
+str_to_datetime(const char *str, size_t length, MYSQL_TIME *l_time,
                 ulonglong flags, MYSQL_TIME_STATUS *status)
 {
   const char *end=str+length, *pos;
@@ -456,7 +457,7 @@ err:
      TRUE  on error
 */
 
-my_bool str_to_time(const char *str, uint length, MYSQL_TIME *l_time,
+my_bool str_to_time(const char *str, size_t length, MYSQL_TIME *l_time,
                     ulonglong fuzzydate, MYSQL_TIME_STATUS *status)
 {
   ulong date[5];
@@ -1425,31 +1426,3 @@ double TIME_to_double(const MYSQL_TIME *my_time)
   d+= my_time->second_part/(double)TIME_SECOND_PART_FACTOR;
   return my_time->neg ? -d : d;
 }
-
-longlong pack_time(const MYSQL_TIME *my_time)
-{
-  return  ((((((my_time->year     * 13ULL +
-               my_time->month)    * 32ULL +
-               my_time->day)      * 24ULL +
-               my_time->hour)     * 60ULL +
-               my_time->minute)   * 60ULL +
-               my_time->second)   * 1000000ULL +
-               my_time->second_part) * (my_time->neg ? -1 : 1);
-}
-
-#define get_one(WHERE, FACTOR) WHERE= (ulong)(packed % FACTOR); packed/= FACTOR
-
-MYSQL_TIME *unpack_time(longlong packed, MYSQL_TIME *my_time)
-{
-  if ((my_time->neg= packed < 0))
-    packed= -packed;
-  get_one(my_time->second_part, 1000000ULL);
-  get_one(my_time->second,           60U);
-  get_one(my_time->minute,           60U);
-  get_one(my_time->hour,             24U);
-  get_one(my_time->day,              32U);
-  get_one(my_time->month,            13U);
-  my_time->year= (uint)packed;
-  my_time->time_type= MYSQL_TIMESTAMP_DATETIME;
-  return my_time;
-}
